
/*!
	@file	CGLCell.h
	@brief	GL cell class
*/

#include <QDebug>
#include <iostream>
#include "CGLCell.h"
#include "font_uv_table.h"

namespace ddv {


CGLCell::CGLCell() {
	m_quadObj = NULL;
}

inline float3 toWorldPosition( float3 f3 ) {
	return float3( f3.x - IMAGEWIDTH_CENTER
		, - f3.y + IMAGEHEIGHT_CENTER, - f3.z );
};

bool CGLCell::drawCell( CCell *pcell ) {
	static const GLfloat red[] = { 0.8, 0.2, 0.2, 1.0 };

	glPushMatrix();
	glEnable( GL_LIGHTING );
	glEnable( GL_CULL_FACE );

	float3 f3 = toWorldPosition( *pcell );
	glTranslatef( f3.x, f3.y, f3.z );
	glMaterialfv( GL_FRONT, GL_DIFFUSE, red );
	GLUquadricObj *quadObj = getQuadObj();
	if ( !quadObj ) {
		glPopMatrix();
		return false;
	}
	gluQuadricDrawStyle( quadObj, GLU_FILL );
	gluQuadricNormals( quadObj, GLU_SMOOTH );
	gluSphere( quadObj, 0.6, 20, 20 );
	glPopMatrix();
	return true;
}

bool CGLCell::drawLineage( CCell *pcell, CCell *pcellParent ) {

	glPushMatrix();
	glEnable( GL_LIGHTING );
	glEnable( GL_CULL_FACE );

	float3 f3Parent = toWorldPosition( *pcellParent );
	float3 f3Daughter = toWorldPosition( *pcell );
	float3 f3ToParent = f3Parent - f3Daughter;
	float3 f3Cross = float3( - f3ToParent.z, 0.0f, f3ToParent.x );
	float fRad = 180.0f / PI * asin( f3Cross.abs()
		/ f3ToParent.abs() * float3( 0.0f, 1.0f, 0.0f ).abs() );
	if ( f3ToParent.y > 0.0f && f3ToParent.z >= 0
		|| f3ToParent.y > 0 && f3ToParent.z < 0 ) {
		fRad *= -1;
	}
	float3 f3Center = ( f3Daughter + f3Parent ) / 2.0f;
	float fT, fLength = f3ToParent.abs() / 2.0f;

	glTranslatef( f3Center.x, f3Center.y, f3Center.z );
	glRotatef( fRad, f3Cross.x, f3Cross.y, f3Cross.z );

	static GLfloat pglfColor[ 4 ] = { 0.9f, 0.9f, 0.9f, 1.0f };
	glMaterialfv( GL_FRONT, GL_DIFFUSE, pglfColor );

	glBegin( GL_QUAD_STRIP );
		for ( int i = 0; i < 7; ++ i ) {
			fT = i * 2.0f * PI / 6.0f;
			glNormal3f( cos( fT ), 0.0f, sin( fT ) );
			glVertex3f( cos( fT ) * 0.1f, - fLength, sin( fT ) * 0.1f );
			glVertex3f( cos( fT ) * 0.1f, fLength, sin( fT ) * 0.1f );
		}
	glEnd();
	glPopMatrix();
	return true;
}

bool CGLCell::drawName( CCell *pcell, matrix *pmatToCameraRotate ) {

	glPushMatrix();
	glDisable( GL_LIGHTING );
	glEnable( GL_BLEND );
	glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
	glEnable( GL_TEXTURE_2D );

	float3 f3 = toWorldPosition( *pcell );
	glTranslatef( f3.x, f3.y, f3.z );
	glMultMatrixf( ( float * ) pmatToCameraRotate );	// billboard rotate
	glTranslatef( 2.0f, 2.0f, 0.0f );					// offset

	glBindTexture( GL_TEXTURE_2D, m_nTextures[ 0 ] );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );

	#define FONTWIDTH 1.0f
	#define FONTHEIGHT 1.5f
	string sName = pcell->getName();
	float x, *pfUV;
	for ( unsigned int i = 0; i < sName.size(); ++ i ) {
		pfUV = g_getUVCoords( sName.c_str()[ i ] );
		if ( !pfUV ) {
			continue;
		}
		x = i * FONTWIDTH;
		glBegin( GL_TRIANGLE_STRIP );
			glTexCoord2f( *pfUV, *( pfUV + 1 ) );
			glVertex3f( x, 0.0f, 0.0f );
			glTexCoord2f( *( pfUV + 2 ), *( pfUV + 1 ) );
			glVertex3f( x + FONTWIDTH, 0.0f, 0.0f );
			glTexCoord2f( *pfUV, *( pfUV + 3 ) );
			glVertex3f( x, FONTHEIGHT, 0.0f );
			glTexCoord2f( *( pfUV + 2 ), *( pfUV + 3 ) );
			glVertex3f( x + FONTWIDTH, FONTHEIGHT, 0.0f );
		glEnd();
	}

	glDisable( GL_TEXTURE_2D );
	glPopMatrix();
	return true;
}

float CGLCell::m_pfOutlineTable[] = {
	0.0f, SCALEUNIT, -SCALEUNIT, SCALEUNIT, 0.0f, 0.0f, 0.0f, 0.0f
	, 0.0f, 0.0f, -SCALEUNIT, 0.0f, -SCALEUNIT, -SCALEUNIT, 0.0f, 0.0f
	, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, -SCALEUNIT, SCALEUNIT, -SCALEUNIT
	, SCALEUNIT, SCALEUNIT, 0.0f, 0.0f, 0.0f, 0.0f, SCALEUNIT, 0.0f };

bool CGLCell::drawOutline( CCell *pcell ) {
  //	glPushMatrix();
  cerr << "1: " << pcell->getName() << endl;
	glDisable( GL_LIGHTING );
	glColor3f( 0.8f, 0.8f, 0.2f );
	for ( vector< COutline >::iterator it = pcell->getOutlines()->begin()
		//		; it != pcell->getOutlines()->end(); ++it ) {
		; it != pcell->getOutlines()->end(); ++it ) {
	  cerr << "2: " << pcell->getName() << endl;
		glBegin( GL_LINE_LOOP );
		char *pcOutline = ( char * ) it->getOutline().c_str();
		float fx = it->x * SCALEUNIT - IMAGEWIDTH_CENTER;
		float fy = - ( it->y * SCALEUNIT - IMAGEHEIGHT_CENTER );
		float fz = - ( it->z * FOCALPLANE_DISTANCE /*- FOCALPLANE_CENTER*/ );
		glVertex3f( fx, fy, fz );
		int nIndex, nCurrent, nPrev = pcOutline[ 0 ] - '0';
		for ( unsigned int i = 0; i < strlen( pcOutline ); ++ i ) {
			nCurrent = pcOutline[ i ] - '0';
			nIndex = nCurrent * 8 + nPrev * 2;
			if ( m_pfOutlineTable[ nIndex ] != 0.0f
				|| m_pfOutlineTable[ nIndex + 1 ] != 0.0f ) {
				fx += m_pfOutlineTable[ nIndex ];
				fy += m_pfOutlineTable[ nIndex + 1 ];
			glVertex3f( fx, fy, fz );
			}
			nPrev = nCurrent;
		}
		glEnd();
	}
	//	glPopMatrix();
	return true;
}

} // namespace ddv

