
/*!
	@file	CQDataset_io.h
	@brief	dataset input/output class
*/

#include <sstream>
#include <QFile>
#include <QTextStream>
#include <cstdlib>
#include <iostream>
#include "CQDataset_io.h"
#include "CCell_io.h"

namespace ddv {

bool CQDataset_io::loadLineage( const char *pcLineageFile ) {

	setCellNameExists( false );

	QFile qfLineage( pcLineageFile );
	if ( !qfLineage.open( QIODevice::ReadOnly | QIODevice::Text ) ) {
		clear();
		m_sErrorString = "load " + string( pcLineageFile ) + " failed.";
		return false;
	}
	QTextStream qtsLineage( &qfLineage );

	QString qsLine;
	string sID, sName;
	int nTimePoint, nParentID, nID;
	float fx, fy, fz, fz_average = 0;
	int nLine = 0, cell_count = 0;
	CCell_io cell_io, *parent_cell, *pcell_io;
	stringstream ssLine( qsLine.toStdString() );
	m_sErrorString = "";
	bool bWithName = true;
	while ( !qtsLineage.atEnd() ) {
		++ nLine;
		qsLine = qtsLineage.readLine();
		ssLine.clear();
		ssLine.str( qsLine.toStdString() );
		ssLine >> nTimePoint >> nParentID >> sID
			>> fx >> fy >> fz >> sName;
		nID = atoi( sID.c_str() );
		if ( !sName.length() ) {
			bWithName = false;
		}

		if ( ( pcell_io = ( CCell_io * ) getCell( nID ) ) ) {
			if ( pcell_io->x != fx || pcell_io->y != fy
				|| pcell_io->z != fz || pcell_io->t != nTimePoint ) {
				clear();
				m_sErrorString
					= "file:" + string( pcLineageFile ) + " is broken.";
				qfLineage.close();
				return false;
			}
			pcell_io->setName( sName );
			continue;
		}
		cell_io.set( nID, nParentID, sName, fx, fy, fz, nTimePoint );
		fz_average += fz;
		++ cell_count;
		if ( nParentID ) {
			parent_cell = ( CCell_io * ) getCell( nParentID );
			parent_cell->addDaughter( nID );
		}
		appendCell( cell_io, nTimePoint );
	}
	if ( !m_fAverageFocalPlane ) {
		m_fAverageFocalPlane = fz_average / cell_count;
	}

	qfLineage.close();
	setLineageExists( true );
	setCellNameExists( bWithName );
	return true;
}

bool CQDataset_io::loadROI( const char *pcROIFile ) {

	setROIExists( false );
	setLineageExists( false );

	QFile qfROI( pcROIFile );
	if ( !qfROI.open( QIODevice::ReadOnly | QIODevice::Text ) ) {
		clear();
		m_sErrorString = "load " + string( pcROIFile ) + " failed.";
		return false;
	}
	QTextStream qtsROI( &qfROI );

	QString qsLine;
	string sType, sID, sOutline;
	int nTimePoint, nParentID, nID;
	float fx, fy, fz, fz_average = 0;
	int nx, ny, nz, nt;
	int nLine = 1, cell_count = 0;
	CCell_io cell_io, *parent_cell;

	bool bError = false;
	qsLine = qtsROI.readLine();
	stringstream ssLine( qsLine.toStdString() );

	m_sErrorString = "";
	while (ssLine >> sType) {
	  if (sType == "G") {
	    if (cell_count != 0) {
	      appendCell(cell_io, nTimePoint);
	    }
	    cell_io.clear();
	    ssLine >> nTimePoint >> nParentID >> sID >> fx >> fy >> fz;
	    nID = atoi(sID.c_str());
	    cell_io.set(nID, nParentID, "", fx, fy, fz, nTimePoint);
	    fz_average += fz;
	    ++cell_count;
	    if (nParentID) {
	      parent_cell = (CCell_io *) getCell(nParentID);
	      parent_cell->addDaughter(nID);
	    }
	    ++nLine;
	    qsLine = qtsROI.readLine();
	    ssLine.clear();
	    ssLine.str(qsLine.toStdString());
	  } else if (sType == "R") {
	    ssLine >> nx >> ny >> nz >> nt >> sOutline;
	    cell_io.appendOutline(nx, ny, nz, sOutline);
	    ++nLine;
	    qsLine = qtsROI.readLine();
	    ssLine.clear();
	    ssLine.str( qsLine.toStdString() );
	  }
	}
	//	cell_io.appendOutline(nx, ny, nz, sOutline);
	appendCell(cell_io, nTimePoint);


	/*
	ssLine >> sType;
	m_sErrorString = "";
	if ( sType != "G" ) {
		clear();
		m_sErrorString = "file:" + string( pcROIFile ) + " is broken.";
		return false;
	}
	do {
		cell_io.clear();
		ssLine >> nTimePoint >> nParentID >> sID >> fx >> fy >> fz;
		nID = atoi( sID.c_str() );
		cell_io.set( nID, nParentID, "", fx, fy, fz, nTimePoint );
		fz_average += fz;
		++ cell_count;
		if ( nParentID ) {
			parent_cell = ( CCell_io * ) getCell( nParentID );
			parent_cell->addDaughter( nID );
		}
		if ( qfROI.atEnd() ) {
			bError = true;
			break;
		}
		++ nLine;
		qsLine = qtsROI.readLine();
		ssLine.clear();
		ssLine.str( qsLine.toStdString() );
		ssLine >> sType;
		if ( sType != "R" ) {
			bError = true;
			break;
		}
		do {
			ssLine >> nx >> ny >> nz >> nt >> sOutline;
			cell_io.appendOutline( nx, ny, nz, sOutline );
			if ( qfROI.atEnd() ) {
				break;
			}
			++ nLine;
			qsLine = qtsROI.readLine();
			ssLine.clear();
			ssLine.str( qsLine.toStdString() );
			ssLine >> sType;
		} while ( sType == "R" );
		std::cerr << nTimePoint << std::endl;
		appendCell( cell_io, nTimePoint );
		if ( qfROI.atEnd() ) {
			break;
		}
	} while ( sType == "G" );
	*/


	if ( bError || !qfROI.atEnd() ) {
		if ( !m_sErrorString.length() ) {
			clear();
			m_sErrorString = "file:" + string( pcROIFile ) + " is broken.";
		} else {
			clear();
		}
		qfROI.close();
		return false;
	}
	m_fAverageFocalPlane = fz_average / cell_count;

	qfROI.close();
	setLineageExists( false );
	setROIExists( true );
	return true;
}

bool CQDataset_io::loadLineageAndROI(
	const char *pcLineageFile, const char *pcROIFile ) {

	bool bResult;
	if ( !( bResult = loadROI( pcROIFile ) ) ) {
		return bResult;
	}
	return loadLineage( pcLineageFile );
}

bool CQDataset_io::appendCell( CCell &cell, int nTimePoint ) {
  std::cerr << nTimePoint << std::endl;
	m_cells.push_back( cell );
	if ( m_indexbyTimePoint.size() < ( unsigned int ) nTimePoint ) {
		vector< CCell * > array;
		m_indexbyTimePoint.push_back( array );
	}
	m_indexbyTimePoint[ nTimePoint - 1 ].push_back( &*( m_cells.rbegin() ) );
	return true;
}

} // namespace ddv

