
/*!
	@file	CQViewer3D.cpp
	@brief	provide 3D cell viewing
*/

#include <Qt>
#include <math.h>
#include <iostream>
#include "constant.h"
#include "CQViewer3D.h"
#include "font_texture.h"

namespace ddv {

void CQViewer3D::clear() {
	m_bMouseLeftPressed = false;
	m_bMouseMiddlePressed = false;
	m_bMouseRightPressed = false;
	m_pcdataset = NULL;

	m_bOutlineEnabled = false;
	m_bNameEnabled = false;
	m_bTraceEnabled = false;
	m_bCenterEnabled = false;

	m_nTextures[ 0 ] = 0;

	resetView();
}

void CQViewer3D::resetView() {
	m_f3CameraTranslate = float3( 0.0f, 0.0f, 10.0f );
	m_qCameraRotate = quaternionRotationAxis(
		float3( 1.0f, 0.0f, 0.0f ), 0.5f );
	m_matCameraRotate = matrix( m_qCameraRotate );
	m_f3CameraScale = float3( 0.08f, 0.08f, 0.08f );
	if ( m_pcdataset ) {
		m_f3ViewPointTranslate = float3( 0.0f, 0.0f
			, - m_pcdataset->getAverageFocalPlane() );
	} else {
		m_f3ViewPointTranslate = float3( 0.0f, 0.0f, 0.0f );
	}
}

void CQViewer3D::initializeGL() {
	qglClearColor( QColor( 0, 0, 0 ) );
	glEnable( GL_DEPTH_TEST );
	glEnable( GL_CULL_FACE );
	glCullFace( GL_BACK );

	glEnable( GL_BLEND );
	glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );

	GLfloat pglfLightPos[] = { 10.0, 10.0, 200.0, 1.0 };
	GLfloat pglfLightColor[] = { 1.0, 1.0, 1.0, 1.0 };
	glLightModeli( GL_LIGHT_MODEL_LOCAL_VIEWER, 1 );
	glLightfv( GL_LIGHT0, GL_POSITION, pglfLightPos );
	glLightfv( GL_LIGHT0, GL_DIFFUSE, pglfLightColor );
	glLightf( GL_LIGHT0, GL_CONSTANT_ATTENUATION, 0.1 );
	glLightf( GL_LIGHT0, GL_LINEAR_ATTENUATION, 0.05 );
	glEnable( GL_LIGHT0 );
	glEnable( GL_LIGHTING );
	resizeGL( width(), height() );
	initializeTextures();
}

void CQViewer3D::initializeTextures() {
	m_qimage = QImage( 128, 128, QImage::Format_ARGB32 );
	for ( int y = 0; y < 128; ++ y ) {
		for ( int x = 0; x < 128; ++ x ) {
			m_qimage.setPixel( x, y
				, ( g_nFontTexture[ y * 128 * 2 + x * 2 ] << 24 )
				| ( g_nFontTexture[ y * 128 * 2 + x * 2 + 1 ] << 8 ) );
		}
	}
	m_nTextures[ 0 ] = bindTexture( m_qimage, GL_TEXTURE_2D, GL_RGBA );
	m_cglcell.setTextures( 0, m_nTextures[ 0 ] );
}

void CQViewer3D::resizeGL( int nWidth, int nHeight ) {
	glViewport( 0, 0, nWidth, nHeight );
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective(
		30.0f, ( double ) nWidth / ( double ) nHeight, 1.0f, 200.0 );
}

void CQViewer3D::paintGL() {
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	draw( GL_RENDER );
}

void CQViewer3D::slotTimePointChanged( int nTimePoint ) {
	if ( m_nTimePoint != nTimePoint ) {
		m_nTimePoint = nTimePoint;
		emit timePointChanged( nTimePoint );
		repaint();
	}
}

list< CCell * > CQViewer3D::sort( vector< CCell * > *punsorted_cells ) {

	matrix m =
		matrix_translate( m_f3ViewPointTranslate )
		* matrix_scale( float3( 1.0f / m_f3CameraScale.x
			, 1.0f / m_f3CameraScale.y, 1.0f / m_f3CameraScale.z ) )
		* m_matCameraRotate
		* matrix_translate( m_f3CameraTranslate );
	float3 f3CameraPos = m * float3( 0.0f, 0.0f, 0.0f );

	list< pair< float, CCell * > > cells;
	float distance;
	float3 f3CellPos;
	for ( vector< CCell * >::iterator it = punsorted_cells->begin()
		; it != punsorted_cells->end(); ++ it ) {
		f3CellPos = float3( ( *it )->x - IMAGEWIDTH_CENTER
			, - ( ( *it )->y - IMAGEHEIGHT_CENTER ), - ( *it )->z );
		distance = ( f3CellPos - f3CameraPos ).abs();
		list< pair< float, CCell * > >::iterator itSort = cells.begin();
		for ( ; itSort != cells.end(); ++ itSort ) {
			if ( itSort->first < distance ) {
				break;
			}
		}
		if ( itSort == cells.end() ) {
			cells.push_back( pair< float, CCell * >( distance, *it ) );
		} else {
			cells.insert( itSort, pair< float, CCell * >( distance, *it ) );
		}
	}

	list< CCell * > sorted_cells;
	for ( list< pair< float, CCell * > >::iterator it = cells.begin()
		; it != cells.end(); ++ it ) {
		sorted_cells.push_back( it->second );
	}
	return sorted_cells;
}

void CQViewer3D::draw( GLenum glnMode ) {
	glPushMatrix();

	glTranslatef( - m_f3CameraTranslate.x, - m_f3CameraTranslate.y
		, - m_f3CameraTranslate.z );
	glMultMatrixf( ( float * ) ( &m_matCameraRotate ) );
	glScalef( m_f3CameraScale.x, m_f3CameraScale.y, m_f3CameraScale.z );
	glTranslatef( - m_f3ViewPointTranslate.x, - m_f3ViewPointTranslate.y
		, - m_f3ViewPointTranslate.z );

	matrix matToCameraRotate = m_matCameraRotate.transpose();
	vector< CCell * > drawnCells;
	if ( glnMode == GL_SELECT ) {
	} else if ( m_pcdataset ) {
		list< CCell * > cells = sort( m_pcdataset->getCellAt( m_nTimePoint ) );
		for ( list< CCell * >::iterator it = cells.begin()
			; it != cells.end(); ++ it ) {
			if ( m_pcdataset->isLineageExists() ) {
				if ( m_bCenterEnabled ) {
					m_cglcell.drawCell( *it );
				}
				if ( m_bTraceEnabled ) {
					drawLineage(*it, &drawnCells );
				}
			}
			if ( m_pcdataset->isROIExists() && m_bOutlineEnabled ) {
			  //			  string sName = (*it)->getName();
			  //			  cerr <<"ID: " << sName << endl;
				m_cglcell.drawOutline( *it );
			}
		}
		for ( list< CCell * >::iterator it = cells.begin()
			; it != cells.end(); ++ it ) {
			if ( m_pcdataset->isCellNameExists() && m_bNameEnabled ) {
				m_cglcell.drawName( *it, &matToCameraRotate );
			}
		}
	}
	glFlush();
	glPopMatrix();
}

void CQViewer3D::drawLineage(CCell *pcell
	, vector< CCell * > *pdrawnCells ) {
	CCell *pcellParent = m_pcdataset->getCell( pcell->getParentID() );
	if ( pcellParent ) {
	  vector< CCell * >::iterator it = pdrawnCells->begin();
	  for ( ; it != pdrawnCells->end(); ++ it ) {
	    if ( *it == pcell ) {
	      break;
	    }
	  }
	  if ( it == pdrawnCells->end() ) {
	    drawLineage(pcellParent, pdrawnCells );
	  }
	  m_cglcell.drawLineage( pcellParent, pcell );
	}
	pdrawnCells->push_back( pcell );
}

void CQViewer3D::mousePressEvent( QMouseEvent *pqmouseevent ) {
	if ( pqmouseevent->buttons() & Qt::LeftButton ) {
		m_int2MouseLeftStart.x = pqmouseevent->x();
		m_int2MouseLeftStart.y = pqmouseevent->y();
		m_int2MousePos = m_int2MouseLeftStart;
		m_bMouseLeftPressed = true;
	}
	if ( pqmouseevent->buttons() & Qt::MidButton ) {
		m_int2MouseMiddleStart.x = pqmouseevent->x();
		m_int2MouseMiddleStart.y = pqmouseevent->y();
		m_int2MousePos = m_int2MouseMiddleStart;
		m_bMouseMiddlePressed = true;
	}
	if ( pqmouseevent->buttons() & Qt::RightButton ) {
		m_int2MouseRightStart.x = pqmouseevent->x();
		m_int2MouseRightStart.y = pqmouseevent->y();
		m_int2MousePos = m_int2MouseRightStart;
		m_bMouseRightPressed = true;
	}
}

void CQViewer3D::mouseReleaseEvent( QMouseEvent *pqmouseevent ) {
	if ( !( pqmouseevent->buttons() & Qt::LeftButton ) ) {
		m_bMouseLeftPressed = false;
	}
	if ( !( pqmouseevent->buttons() & Qt::MidButton ) ) {
		m_bMouseMiddlePressed = false;
	}
	if ( !( pqmouseevent->buttons() & Qt::RightButton ) ) {
		m_bMouseRightPressed = false;
	}
}

void CQViewer3D::mouseMoveEvent( QMouseEvent *pqmouseevent ) {
	m_int2MousePosOld = m_int2MousePos;
	m_int2MousePos.x = pqmouseevent->x();
	m_int2MousePos.y = pqmouseevent->y();
	if ( m_bMouseLeftPressed ) {
		mouseLeftButton_dragged( pqmouseevent );
	}
	if ( m_bMouseMiddlePressed ) {
		mouseMiddleButton_dragged( pqmouseevent );
	}
	if ( m_bMouseRightPressed ) {
		mouseRightButton_dragged( pqmouseevent );
	}
}

void CQViewer3D::mouseLeftButton_dragged( QMouseEvent *pqmouseevent ) {
	float f = 2.0f * PI
		* ( m_int2MousePosOld.x - m_int2MousePos.x ) / ( float ) IMAGEHEIGHT;
	quaternion qx, qz;
	if ( f ) {
		qz = quaternionRotationAxis( float3( 0.0f, 0.0f, 1.0f ), f );
	}
	f = 2.0f * PI
		* ( m_int2MousePosOld.y - m_int2MousePos.y ) / ( float ) IMAGEWIDTH;
	if ( f ) {
		qx = quaternionRotationAxis( float3( 1.0f, 0.0f, 0.0f ), f );
	}
	m_qCameraRotate = qz * m_qCameraRotate * qx;
	m_matCameraRotate = matrix( m_qCameraRotate );
	repaint();
}

void CQViewer3D::mouseMiddleButton_dragged( QMouseEvent *pqmouseevent ) {
	m_f3ViewPointTranslate -= matrix( m_qCameraRotate ) * float3(
		- ( m_int2MousePosOld.x - m_int2MousePos.x ) / 20.0f
		, ( m_int2MousePosOld.y - m_int2MousePos.y ) / 20.0f
		, 0.0f );
	repaint();
}

void CQViewer3D::mouseRightButton_dragged( QMouseEvent *pqmouseevent ) {
	m_f3CameraTranslate.z -= ( m_int2MousePos.x - m_int2MousePosOld.x ) / 20.0f;
	char c[ 1024 ];
	sprintf( c, "%f", m_f3CameraTranslate.z );
	showMessage( QString( "distance: " ) + QString( c ), 10000 );
	repaint();
}

void CQViewer3D::wheelEvent( QWheelEvent *pqwheelevent ) {
	if ( !m_pcdataset ) {
		return;
	}
	if ( pqwheelevent->delta() > 0 ) {
		if ( m_nTimePoint > m_pcdataset->getMinTimePoint() ) {
			emit timePointChanged( m_nTimePoint - 1 );
		}
	} else {
		if ( m_nTimePoint < m_pcdataset->getMaxTimePoint() ) {
			emit timePointChanged( m_nTimePoint + 1 );
		}
	}
	repaint();
}

void CQViewer3D::slotActionViewOutline_toggled( bool bEnabled ) {
	m_bOutlineEnabled = bEnabled;
	repaint();
}

void CQViewer3D::slotActionViewName_toggled( bool bEnabled ) {
	m_bNameEnabled = bEnabled;
	repaint();
}

void CQViewer3D::slotActionViewTrace_toggled( bool bEnabled ) {
	m_bTraceEnabled = bEnabled;
	repaint();
}

void CQViewer3D::slotActionViewCenter_toggled( bool bEnabled ) {
	m_bCenterEnabled = bEnabled;
	repaint();
}

void CQViewer3D::slotResetView() {
	resetView();
	repaint();
}

} // namespace ddv

