
/*!
	@file	CQViewer3D.h
	@brief	provide 3D cell viewing
*/

#ifndef __H_CQVIEWER3D__
#define __H_CQVIEWER3D__

#include <list>
#include <vector>
#include <QGLWidget>
#include <QMouseEvent>
#include "CDataset.h"
#include "CCell.h"
#include "CGLCell.h"
#include "int2.h"
#include "float3.h"
#include "matrix.h"
#include "quaternion.h"

namespace ddv {

class CQViewer3D: public QGLWidget {

	Q_OBJECT

public:
	CQViewer3D( QWidget *parent ): QGLWidget( parent )
		, m_pcdataset( NULL ) {
		setMouseTracking( true );
		setFixedSize( 600, 600 );
		clear();
	};
	~CQViewer3D() {
		if ( m_nTextures[ 0 ] ) {
			deleteTexture( m_nTextures[ 0 ] );
			m_nTextures[ 0 ] = 0;
		}
	};

	void clear();

	inline void setDataset( CDataset *pcdataset ) {
		m_pcdataset = pcdataset;
		resetView();
	};

	void resetView();

protected:
	void initializeGL();
	void initializeTextures();
	void resizeGL( int intWidth, int intHeight );
	void paintGL();

	list< CCell * > sort( vector< CCell * > *punsorted_cells );

signals:
	void showMessage( const QString &message, int timeout = 0 );
	void timePointChanged( int nTimePoint );

public slots:
	void slotActionViewOutline_toggled( bool bEnabled );
	void slotActionViewName_toggled( bool bEnabled );
	void slotActionViewTrace_toggled( bool bEnabled );
	void slotActionViewCenter_toggled( bool bEnabled );
	void slotTimePointChanged( int nTimePoint );
	void slotResetView();

protected:
	void draw( GLenum glnMode );
	void drawLineage(CCell *pcell, vector< CCell * > *pdrawnCell );

	void mousePressEvent( QMouseEvent *pqmouseevent );
	void mouseReleaseEvent( QMouseEvent *pqmouseevent );
	void mouseMoveEvent( QMouseEvent *pqmouseevent );
	void mouseLeftButton_dragged( QMouseEvent *pqmouseevent );
	void mouseMiddleButton_dragged( QMouseEvent *pqmouseevent );
	void mouseRightButton_dragged( QMouseEvent *pqmouseevent );
	void wheelEvent( QWheelEvent *pqwheelevent );

	int m_nTimePoint;
	int2 m_int2MousePos;
	int2 m_int2MousePosOld;
	int2 m_int2MouseLeftStart;
	int2 m_int2MouseMiddleStart;
	int2 m_int2MouseRightStart;
	bool m_bMouseLeftPressed;
	bool m_bMouseMiddlePressed;
	bool m_bMouseRightPressed;

	bool m_bOutlineEnabled;
	bool m_bNameEnabled;
	bool m_bTraceEnabled;
	bool m_bCenterEnabled;

	float3 m_f3CameraTranslate;
	float3 m_f3CameraRotate;
	matrix m_matCameraRotate;
	quaternion m_qCameraRotate;
	float3 m_f3CameraScale;
	float3 m_f3ViewPointTranslate;
	quaternion m_qBegin;

	CDataset *m_pcdataset;
	CGLCell m_cglcell;
	QImage m_qimage;
	GLuint m_nTextures[ 1 ];
};

} // namespace ddv

#endif // __H_CQVIEWER3D__

