
#ifndef __H_MATRIX__
#define __H_MATRIX__

#include "float3.h"
#include "quaternion.h"
#include <QDebug>

namespace ddv {

typedef struct matrix {
public:
	matrix() {};
	matrix( float a, float b, float c, float d
		, float e, float f, float g, float h
		, float i, float j, float k, float l
		, float m, float n, float o, float p )
		: _11( a ), _12( b ), _13( c ), _14( d )
		, _21( e ), _22( f ), _23( g ), _24( h )
		, _31( i ), _32( j ), _33( k ), _34( l )
		, _41( m ), _42( n ), _43( o ), _44( p ) {};
	matrix( quaternion q ) {
		float x, y, z, t, x2, y2, z2, xy, yz, zx, tx, ty, tz;
		x = q.x;
		y = q.y;
		z = q.z;
		t = q.t;
		x2 = 2.0f * x * x;
		y2 = 2.0f * y * y;
		z2 = 2.0f * z * z;
		xy = 2.0f * x * y;
		yz = 2.0f * y * z;
		zx = 2.0f * z * x;
		tx = 2.0f * t * x;
		ty = 2.0f * t * y;
		tz = 2.0f * t * z;
		_11 = 1.0f - y2 - z2;
		_12 = xy - tz;
		_13 = zx + ty;
		_14 = 0.0f;
		_21 = xy + tz;
		_22 = 1.0f - z2 - x2;
		_23 = yz - tx;
		_24 = 0.0f;
		_31 = zx - ty;
		_32 = yz + tx;
		_33 = 1.0f - x2 - y2;
		_34 = 0.0f;
		_41 = 0.0f;
		_42 = 0.0f;
		_43 = 0.0f;
		_44 = 1.0f;
	};

	inline float3 operator * ( float3 f3 ) {
		return float3(
			  _11 * f3.x + _12 * f3.y + _13 * f3.z + _14
			, _21 * f3.x + _22 * f3.y + _23 * f3.z + _24
			, _31 * f3.x + _32 * f3.y + _33 * f3.z + _34 );
	};
	inline matrix operator * ( matrix m ) {
		return matrix(
			  _11 * m._11 + _12 * m._21 + _13 * m._31 + _14 * m._41
			, _11 * m._12 + _12 * m._22 + _13 * m._32 + _14 * m._42
			, _11 * m._13 + _12 * m._23 + _13 * m._33 + _14 * m._43
			, _11 * m._14 + _12 * m._24 + _13 * m._34 + _14 * m._44

			, _21 * m._11 + _22 * m._21 + _23 * m._31 + _24 * m._41
			, _21 * m._12 + _22 * m._22 + _23 * m._32 + _24 * m._42
			, _21 * m._13 + _22 * m._23 + _23 * m._33 + _24 * m._43
			, _21 * m._14 + _22 * m._24 + _23 * m._34 + _24 * m._44

			, _31 * m._11 + _32 * m._21 + _33 * m._31 + _34 * m._41
			, _31 * m._12 + _32 * m._22 + _33 * m._32 + _34 * m._42
			, _31 * m._13 + _32 * m._23 + _33 * m._33 + _34 * m._43
			, _31 * m._14 + _32 * m._24 + _33 * m._34 + _34 * m._44

			, _41 * m._11 + _42 * m._21 + _43 * m._31 + _44 * m._41
			, _41 * m._12 + _42 * m._22 + _43 * m._32 + _44 * m._42
			, _41 * m._13 + _42 * m._23 + _43 * m._33 + _44 * m._43
			, _41 * m._14 + _42 * m._24 + _43 * m._34 + _44 * m._44 );
	};

	inline void identity() {
		_11 = 1.0f; _12 = 0.0f; _13 = 0.0f; _14 = 0.0f;
		_21 = 0.0f; _22 = 1.0f; _23 = 0.0f; _24 = 0.0f;
		_31 = 0.0f; _32 = 0.0f; _33 = 1.0f; _34 = 0.0f;
		_41 = 0.0f; _42 = 0.0f; _43 = 0.0f; _44 = 1.0f;
	};

	inline matrix transpose() {
		return matrix(
			  _11, _21, _31, _41
			, _12, _22, _32, _42
			, _13, _23, _33, _43
			, _14, _24, _34, _44 );
	};

	float _11, _12, _13, _14
		, _21, _22, _23, _24
		, _31, _32, _33, _34
		, _41, _42, _43, _44;
};

inline matrix matrix_translate( float3 f3 ) {
	return matrix(
		  1.0f, 0.0f, 0.0f, f3.x
		, 0.0f, 1.0f, 0.0f, f3.y
		, 0.0f, 0.0f, 1.0f, f3.z
		, 0.0f, 0.0f, 0.0f, 1.0f );
};

inline matrix matrix_rotate_x( float f ) {
	return matrix(
		1.0f, 0.0f, 0.0f, 0.0f
		, 0.0f, cos( f ), -sin( f ), 0.0f
		, 0.0f, sin( f ), cos( f ), 0.0f
		, 0.0f, 0.0f, 0.0f, 1.0f );
};

inline matrix matrix_rotate_y( float f ) {
	return matrix(
		cos( f ), 0.0f, sin( f ), 0.0f
		, 0.0f, 1.0f, 0.0f, 0.0f
		, -sin( f ), 0.0f, cos( f ), 0.0f
		, 0.0f, 0.0f, 0.0f, 1.0f );
};

inline matrix matrix_rotate_z( float f ) {
	return matrix(
		cos( f ), -sin( f ), 0.0f, 0.0f
		, sin( f ), cos( f ), 0.0f, 0.0f
		, 0.0f, 0.0f, 1.0f, 0.0f
		, 0.0f, 0.0f, 0.0f, 1.0f );
};

inline matrix matrix_scale( float3 f3 ) {
	return matrix(
		  f3.x, 0.0f, 0.0f, 0.0f
		, 0.0f, f3.y, 0.0f, 0.0f
		, 0.0f, 0.0f, f3.z, 0.0f
		, 0.0f, 0.0f, 0.0f, 1.0f );
};

} // namespace ddv

#endif // __H_MATRIX__

